#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'requests>=2.19.1', 'python-magic>=0.4.15', 'dateutils>=0.6.6',
    'xmltodict>=0.12.0', 'pyyaml>=5.1.2' ]

setup(
    name='chibi',
    keywords='chibi',
    version='0.11.3',
    description='',
    long_description=readme + '\n\n' + history,
    license="WTFPL",
    author='dem4ply',
    author_email='',
    packages=find_packages(include=['chibi', 'chibi.*']),
    install_requires=requirements,
    dependency_links = [],
    url='https://github.com/dem4ply/chibi',
    zip_safe=False,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: Public Domain',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ] )



import importlib.util
from chibi.file import Chibi_file


__all__ = [ 'Chibi_python' ]


class Chibi_python( Chibi_file ):
    def read( self ):
        value = super().read()
        return value

    def import_( self ):
        name = self.file_name[:-3]
        spec = importlib.util.spec_from_file_location(
            name, str( self.path.inflate ) )
        module = importlib.util.module_from_spec( spec )
        spec.loader.exec_module( module )
        return module



from tests.snippet.files import Test_with_files
from chibi.file import Chibi_file


class Test_chibi_file_check_sum_md5( Test_with_files ):
    def setUp( self ):
        super().setUp()
        self.chibi_file = Chibi_file( self.files[0] )

    def test_should_do_check_sum_empty_file( self ):
        self.assertTrue(
            self.chibi_file.check_sum_md5( '1B2M2Y8AsgTpgAmY7PhCfg==' ) )



from collections import OrderedDict

from .descriptor import Descriptor


class Chibi_object_meta_base():
    _fields = {}


class Chibi_object_meta( type ):
    @classmethod
    def __prepare__( cls, name, bases ):
        return OrderedDict()

    def __new__( cls, clsname, bases, clsdict ):

        fields = [
            k for k, v in clsdict.items() if isinstance( v, Descriptor ) ]
        for name in fields:
            clsdict[name].name = name

        attr_meta = clsdict.pop( 'Meta', None )
        clsobj = super().__new__( cls, clsname, bases, clsdict )

        if attr_meta is None:
            if hasattr( clsobj, '_meta' ):
                attr_meta = clsobj._meta
            else:
                attr_meta = Chibi_object_meta_base()
            clsobj.Meta = attr_meta
        else:
            clsobj.Meta = attr_meta
        meta_base = getattr( clsobj, 'Meta', None )
        clsobj._meta = Chibi_object_meta_base()
        clsobj._meta.__dict__.update( meta_base.__dict__.copy() )
        attr_meta = clsobj._meta

        if hasattr( attr_meta, '_fields' ):
            attr_meta._fields = {}
        attr_meta._fields.update( meta_base._fields )
        attr_meta._fields.update(
            { name: clsdict[ name ] for name in fields } )
        return clsobj


class Chibi_object( metaclass=Chibi_object_meta ):
    """
    Notes
    -----
    la clase de Meta se guarda como una clase interna llamada _meta en la
        instancia
    atributos de la clase _meta
    ---------------------------
        _fields: este dicionario guarda el nombre y la instancia del campo
    """
    class Meta( Chibi_object_meta_base ):
        pass

    def __init__( self, *args, **kargs ):
        for name, field in self._meta._fields.items():
            setattr( self, name, kargs.get( name, field.default ) )


class Chibi_object_iterate_field( Chibi_object ):
    class Meta( Chibi_object.Meta ):
        iterate_field = None

    def __init__( self, *args, **kargs ):
        if not hasattr( self, self.Meta.iterate_field ):
            AttributeError(
                "No found attribute found %s" % self.Meta.iterate_field )
        super().__init__(*args, **kargs )

    def __iter__( self ):
        return iter( getattr( self, self.Meta.iterate_field ) )


__all__ = [Chibi_object, Chibi_object_iterate_field]



fix_values_of_bool = {
    'true': True,
    'false': False,
    '0': False,
    'y': True,
    'n': False,
}


def link_header( header, link_delimiter=',', param_delimiter=';' ):
    """
    Parsea en dicionario la respusta de las cabezera ``link``

    Parameters
    ----------
    header: string
        contenido de la cabezera ``link``
    link_delimiter: Optional[str]
        caracter que separa los enlaces
    param_delimiter: Optional[str]
        caracter que separa el parametro ``rel`` del link

    Returns
    -------
    dict

    Note
    ----
    enlace a la especificacion del header
    `link header <https://tools.ietf.org/html/rfc5988>`_
    """
    result = {}
    links = header.split( link_delimiter )
    for link in links:
        segments = link.split( param_delimiter )
        if len( segments ) < 2:
            continue
        link_part = segments[0].strip()
        rel_part = segments[1].strip().split( '=' )[1][1:-1]
        if not link_part.startswith( '<' ) and not link_part.endswith( '>' ):
            continue
        link_part = link_part[1:-1]
        result[rel_part] = link_part
    return result


def to_bool( value ):
    """
    tranforma el valor en booleana usando una lista fija

    Arguments
    ---------
    value: any

    Returns
    -------
    bool

    Examples
    >>>to_bool( '0' )
    False
    >>>to_bool( 'false' )
    False
    >>>to_bool( '1' )
    True
    >>>to_bool( 'true' )
    True
    >>>to_bool( object() )
    True
    """
    if isinstance( value, str ):
        return fix_values_of_bool.get( value.lower(), bool( value ) )
    return fix_values_of_bool.get( value, bool( value ) )



from chibi.file.snippets import is_file
from tests.snippet.files import Test_with_files


class Test_is_file( Test_with_files ):
    amount_of_files = 3
    amount_of_dirs = 3

    def test_root_should_be_a_false( self ):
        self.assertFalse( is_file( self.root_dir ) )

    def test_all_dirs_list_should_be_false( self ):
        for dir in self.dirs:
            self.assertFalse( is_file( dir ) )

    def test_all_files_liist_should_be_true( self ):
        for file in self.files:
            self.assertTrue( is_file( file ) )



import yaml
from chibi.file import Chibi_file
from chibi.atlas import _wrap


__all__ = [ 'Chibi_yaml' ]


class Chibi_yaml( Chibi_file ):
    def read( self ):
        self.reread()
        result = yaml.load( self.file, Loader=yaml.FullLoader )
        return _wrap( result )

    def write( self, data, is_safe=False ):
        if is_safe:
            super().write( yaml.safe_dump( data ) )
        else:
            super().write( yaml.dump( data ) )



from chibi.file.snippets import stat
from tests.snippet.files import Test_with_files


class Test_stat_to_dir( Test_with_files ):
    amount_of_files = 3
    amount_of_dirs = 3

    def test_stat_to_dir_should_no_raise_a_exception( self ):
        stat( self.root_dir )

    def test_stat_to_dir_should_return_the_group( self ):
        result = stat( self.root_dir )
        self.assertIn( 'group', result )

    def test_stat_to_dir_should_return_the_user( self ):
        result = stat( self.root_dir )
        self.assertIn( 'user', result )


class Test_stat_to_file( Test_with_files ):
    amount_of_files = 3
    amount_of_dirs = 3

    def test_stat_to_file_should_no_raise_a_exception( self ):
        stat( self.files[0] )

    def test_stat_to_file_should_return_the_group( self ):
        result = stat( self.files[0] )
        self.assertIn( 'group', result )

    def test_stat_to_file_should_return_the_user( self ):
        result = stat( self.files[0] )
        self.assertIn( 'user', result )



from . import Chibi_atlas


class Chibi_atlas_multi( Chibi_atlas ):
    def __setitem__( self, name, value ):
        if name in self:
            current = self[ name ]
            if not isinstance( current, list ):
                current = [ current ]
                del self[ name ]
                self[ name ] = current
            self[ name ].append( value )
        else:
            super().__setitem__( name, value )

    def update( self, d ):
        if not isinstance( d, dict ):
            raise NotImplementedError(
                f"no esta implementado el typo {type(d)}" )
        for k, v in d.items():
            self[ k ] = v



import glob
import hashlib
import os
import re
import shutil
import logging

from chibi import b64
from chibi.atlas import Chibi_atlas
from chibi.file.path import Chibi_path
from chibi.snippet import regex


logger = logging.getLogger( "chibi.file.snippets" )


def current_dir():
    """
    regresa el directorio actual de trabajo

    Returns
    =======
    py:class:`chibi.file.Chibi_path`
    """
    logger.warning( "current_dir de snippets siendo usado" )
    return Chibi_path( os.getcwd() )


def file_dir( f ):
    return os.path.dirname( os.path.realpath( f ) )


def base_name( path ):
    return os.path.basename( path )


def cd( to ):
    """
    cambia el directorio actual

    Parameters
    ==========
    to: string
        dirrecion a la que se quiere cambiar

    Returns
    =======
    None
    """
    to = inflate_dir( to )
    os.chdir( to )


def inflate_dir( src ):
    """
    infla la dirrecion para obtner la ruta absoluta

    Parameters
    ==========
    src: string
        direcion que se quiere inflar

    Returns
    =======
    string
    """
    if '~' in src:
        return os.path.expanduser( src )
    else:
        return os.path.abspath( src )


def is_dir( src ):
    """
    revisa si una direcion es un directorio

    Returns
    =======
    bool
    """
    return os.path.isdir( src )


def is_file( src ):
    """
    revisa si una direcion es un archivo

    Returns
    =======
    bool
    """
    return os.path.isfile( src )


def get_file_from_path( src ):
    s = os.path.split( src )
    return s[-1]


def ls( src=None, recursive=False ):
    """
    lo mismo que ls en unix

    Returns
    =======
    iterador of strings
    """
    if src is None:
        src = current_dir()
    else:
        src = Chibi_path( src )
    if not recursive:
        dirs = glob.iglob( src )
        for d in dirs:
            try:
                for name in os.listdir( d ):
                    yield Chibi_path( d ) + name
            except NotADirectoryError:
                yield Chibi_path( d )
    else:
        for r in ls( src ):
            yield Chibi_path( r )
            if is_a_folder( r ):
                for rr in ls( r, recursive=True ):
                    yield Chibi_path( rr )


def find( src=None, search_term=r'.*' ):
    name = re.compile( search_term )
    for f in ls( src, recursive=True ):
        last_part = base_name( f )
        if regex.test( name, last_part ):
            yield f


def find_only_files( src=None, search_term=r'.*' ):
    return (
        f for f in find( src, search_term=search_term )
        if is_a_file( f ) )


def find_only_folders( src=None, search_term=r'.*' ):
    return (
        f for f in find( src, search_term=search_term )
        if is_a_folder( f ) )


def ls_only_files( src=None ):
    """
    lo mismo que ls en unix pero solo archivos

    Returns
    =======
    iterador of strings
    """
    return ( f for f in ls( src ) if is_file( f ) )


def ls_only_dir( src=None ):
    """
    lo mismo que ls en unix pero solo directorios

    Returns
    =======
    iterador of strings
    """
    return ( name for name in ls( src ) if name.is_a_folder )


def mkdir( new_dir, is_ok_exists=True, verbose=True ):
    """
    crea un nuevo directiorio

    Parameters
    ==========
    new_dir: string
        nombre o direcion del nuevo directorio
    is_ok_exists: bool
        define si esta bien is el directorio ya existe
        y no lanza una exception
    """
    try:
        os.makedirs( inflate_dir( new_dir ), )
        if verbose:
            logger.info( f"se creo el directorio '{new_dir}'" )
    except OSError:
        if not is_ok_exists:
            raise


def join( *patch ):
    """
    une los argumentes en una direcion

    Parameters
    ==========
    patch: list of strings

    Returns
    =======
    string
    """
    logger.warning( "join de snippets siendo usado" )
    patch = tuple( str( a ) for a in patch )
    return Chibi_path( os.path.join( *patch ) )


def exists( file_name ):
    """
    revisa si el archivo existe en la ruta

    Parameters
    ==========
    file_name: string

    Returns
    =======
    bool
    """
    return os.path.exists( file_name )


def move( source, dest, verbose=True ):
    """
    mueve archivos

    Parameters
    ==========
    source: string
        ruta del archivo original
    dest: string
        ruta del destine
    verbose: bool

    Returns
    =======
    None
    """
    g = glob.glob( str( source ) )
    dest = str( dest )
    if len( g ) > 1 and not is_a_folder( dest ):
        raise ValueError( "'{}' was expected be a dir".format( dest ) )
    for f in g:
        shutil.move( f, dest )
        if verbose:
            logger.info( f"{f} -> {dest}" )


def is_a_folder( dir ):
    return os.path.isdir( dir )


def is_a_file( f ):
    return os.path.isfile( f )


def copy( source, dest, verbose=True ):
    """
    copia el contenido de un archivo al destino

    Parameters
    ==========
    source: string
        ruta del archivo original
    dest: string
        ruta del destine del nuevo archivo
    verbose: bool

    Returns
    =======
    Non
    """
    g = glob.glob( source )
    if len( g ) > 1 and not is_a_folder( dest ):
        raise ValueError( "'{}' was expected be a dir".format( dest ) )
    for f in g:
        shutil.copy( f, dest )
        if verbose:
            logger.info( f"{f} -> {dest}" )


def copy_folder( source, dest ):
    shutil.copytree( str( source ), str( dest ) )


def ln( source, dest, verbose=True ):
    os.symlink( source, dest )
    if verbose:
        logger.info( f"{source} -> {dest}" )


def _print_verboce_chown( path, old_stat, current_stat ):
    user_change = old_stat.user.name != current_stat.user.name
    group_change = old_stat.group.name != current_stat.group.name

    if user_change or group_change:
        logger.info(
            "el propietario de '{path}' cambio de '{old}' a '{new}'"
            .format(
                path=path, new="{user}:{group}".format(
                    user=current_stat.user.name,
                    group=current_stat.group.name ),
                old="{user}:{group}".format(
                    user=old_stat.user.name, group=old_stat.group.name ) ) )
    else:
        logger.info(
            "el propietario de '{path}' permanece '{old}'"
            .format(
                path=path, new="{user}:{group}".format(
                    user=current_stat.user.name,
                    group=current_stat.group.name ),
                old="{user}:{group}".format(
                    user=old_stat.user.name, group=old_stat.group.name ) ) )


def read_in_chunks( file_name, prop='rb', chunk_size=4096 ):
    with open( file_name, prop ) as f:
        while True:
            current_read = f.read( chunk_size )
            if not current_read:
                break
            yield current_read


def check_sum_md5( file_name, check_sum ):
    md5 = hashlib.md5()
    for chunk in read_in_chunks( file_name ):
        md5.update( chunk )
    md5_bin = md5.digest()
    return b64.encode( md5_bin ) == check_sum


def delete( path ):
    if is_file( path ):
        os.remove( path )
    else:
        shutil.rmtree( path )


def chown(
        *paths, verbose=True, user_name=None, group_name=None,
        recursive=False ):
    from chibi.nix import get_passwd, get_group
    if user_name is None:
        user = None
        uid = -1
    else:
        user = get_passwd( name=user_name )
        uid = user.uid

    if group_name is None:
        group = None
        gid = -1
    else:
        group = get_group( name=group_name )
        gid = group.gid

    for path in paths:
        path = Chibi_path( path )
        old_stat = stat( path )
        os.chown( path, uid, gid )
        current_stat = stat( path )
        if verbose:
            _print_verboce_chown( path, old_stat, current_stat )

        if recursive and path.is_a_folder:
            inner_paths = list( path.ls() )
            chown(
                *inner_paths, user_name=user_name, group_name=group_name,
                verbose=verbose, recursive=True )


def stat( src ):
    from chibi.nix import get_passwd, get_group
    s = os.stat( src )
    result = Chibi_atlas( dict(
        mode=s.st_mode, ino=s.st_ino, dev=s.st_dev,
        nlink=s.st_nlink, size=s.st_size, atime=s.st_atime, mtime=s.st_mtime,
        ctime=s.st_ctime ) )
    result.user = get_passwd( uid=s.st_uid )
    result.group = get_group( gid=s.st_gid )

    return result


def add_extensions( file_name, *extensions ):
    """
    agrega extenciones al nombre del archivo

    Parameters
    ==========
    file_name: str
    extensions: tuple of str

    Returns
    =======
    str

    Examples
    ========
    >>>add_extensions( "image.jpg", "thumbnail" )
    "image.thumbnail.jpg"
    """
    file_name, ext = os.path.splitext( file_name )
    extensions = ".".join( extensions )
    file_name = ".".join( ( file_name, extensions ) )
    return Chibi_path( file_name + ext )


def common_root( *paths ):
    """
    encuentra el la carpeta raiz en comun

    Parameters
    ==========
    paths: tuple of str or Chibi_path

    Returns
    =======
    py:class:`chibi.file.Chibi_path`

    Examples
    ========
    >>>common_root( '/usr/var/log', '/usr/var/security' )
    '/usr/var/'
    """
    return Chibi_path( os.path.commonprefix( paths ) )


def get_relative_path( *paths, root=None ):
    """
    regresa los path relativos

    Parameters
    ==========
    paths: tuple of str or Chibi_path
    root: string

    Returns
    =======
    list of strings
        cuando se manda mas de path
    string
        cuando solo se envia 1 path

    Examples
    ========
    >>>get_relative_path( '/usr/var/log', '/usr/var/security' )
    [ 'log', 'security' ]
    >>>get_relative_path( '/usr/var/log', root='/usr/var/' )
    'log'
    """
    if len( paths ) > 1:
        if root is None:
            root = common_root( *paths )
    else:
        if root is None:
            raise NotImplementedError
    root = inflate_dir( root )

    result = [
        os.path.relpath( inflate_dir( path ), start=root )
        for path in paths ]
    if len( result ) == 1:
        return result[0]
    return result



import json
import xml
from collections import defaultdict

import xmltodict
import yaml

from chibi.snippet.dict import hate_ordered_dict, remove_xml_notatation


class Atlas:
    def __new__( cls, item, *args, **kw  ):
        return _wrap( item )


def loads( string ):
    try:
        return Chibi_atlas( json.loads( string ) )
    except ( json.JSONDecodeError, TypeError ):
        try:
            result = xmltodict.parse( string )
            result = hate_ordered_dict( result )
            result = remove_xml_notatation( result )
            return Chibi_atlas( result )
        except xml.parsers.expat.ExpatError:
            return Chibi_atlas( yaml.safe_load( string ) )


class Chibi_atlas( dict ):
    """
    Clase para crear dicionarios para que sus keys sean leibles como
    atributos de classes
    """

    def __init__( self, *args, **kw ):
        for arg in args:
            if isinstance( arg, dict ):
                for k, v in arg.items():
                    self[ k ] = v
        for k, v in kw.items():
            self[ k ] = v

        # super().__init__( *args, **kw )

    def __getattr__( self, name ):
        try:
            return super().__getattribute__( name )
        except AttributeError as e:
            try:
                return self[ name ]
            except KeyError:
                raise e

    def __setattr__( self, name, value ):
        try:
            if getattr( type( self ), name, False ):
                super().__setattr__( name, value )
            else:
                self[ name ] = _wrap( value )
        except TypeError:
            self[ name ] = _wrap( value )

    def __delattr__( self, name ):
        del self[ name ]

    def __setitem__( self, name, value ):
        super().__setitem__( name, _wrap( value ) )

    def __dir__( self ):
        return list( self.keys() )


class Chibi_atlas_ignore_case( Chibi_atlas ):
    """
    clase que crea chibi atlas que son case insensitive
    """
    def __init__( self, *args, **kw ):
        args_clean = []
        for a in args:
            if isinstance( a, dict ) or hasattr( a, 'items' ):
                args_clean.append( { k.lower(): v for k, v in a.items() } )
        kw = { k.lower(): v for k, v in kw.items() }
        super().__init__( *args_clean, **kw )

    def __getattr__( self, name ):
        name = name.lower()
        return super().__getattr__( name )

    def __getitem__( self, key ):
        key = key.lower()
        return super().__getitem__( key )

    def __setattr__( self, name, value ):
        name = name.lower()
        return super().__setattr__( name, value )

    def __setitem__( self, key, value ):
        key = key.lower()
        return super().__setitem__( key, value )


def _default_factory():
    return Chibi_atlas_default()


class Chibi_atlas_default( defaultdict, Chibi_atlas ):
    """
    chibi atlas que emula `py:class:collections.defaultdict`
    """
    def __init__( self, default_factory=None, *args, **kw ):
        if default_factory is None:
            default_factory = _default_factory
        super().__init__( default_factory, *args, **kw )


class __Chibi_atlas_list( list ):
    def __getitem__( self, index ):
        value = super().__getitem__( index, )
        value = _wrap( value )
        self[ index ] = value
        return value

    def __iter__( self ):
        for i, v in enumerate( super().__iter__() ):
            value = _wrap( v )
            self[ i ] = value
            yield value


def _wrap( val, klass=None ):
    if type( val ) == dict:
        if klass is None:
            return Chibi_atlas( val )
        else:
            return klass( val )
    elif type( val ) == list:
        if klass is None:
            return __Chibi_atlas_list( val )
        else:
            return klass( val )
    return val


yaml.add_representer(
    Chibi_atlas, yaml.representer.SafeRepresenter.represent_dict )


yaml.add_representer(
    __Chibi_atlas_list, yaml.representer.SafeRepresenter.represent_list )



import copy


__all__ = [ 'Pipeline' ]


class Pipeline_manager:
    def __init__( self, *args, **kw ):
        self.children = []

    def __repr__( self ):
        result = "<class {}( children={} )>".format(
            self.__class__, self.children )
        return result

    def __copy__( self ):
        new_node = self.__class__()
        for child in self.children:
            new_node.append( child )
        return new_node

    def append( self, other ):
        self.children.append( other )

    def __add__( self, other ):
        result = copy.copy( self )
        result.children += other.children
        return result

    def __or__( self, other ):
        me = copy.deepcopy(self)
        if isinstance( other, Pipeline_manager ):
            return me + other

        me.append(other)
        return me

    def to_dict( self ):
        result = {
            'children': [ child for child in self.children ]
        }
        return result

    def run( self, obj ):
        result = obj
        for node in self.children:
            if isinstance( node, type ):
                node = node()
            result = node.process( result )
        return result


class Pipeline_meta( type ):
    def __or__( cls, other ):
        return Pipeline_manager() | cls | other


class Pipeline( metaclass=Pipeline_meta ):
    FUN = None

    def __init__( self, *args, fun=None, **kw ):
        if fun is None:
            self._fun = self.FUN
        else:
            self._fun = fun

    def run( self, obj, *arg, **kw ):
        if not self._fun:
            raise NotImplementedError
        else:
            return self._fun( obj )

    def __or__( self, other ):
        return Pipeline_manager() | self | other



from chibi.parser import to_bool
from unittest import TestCase


class Test_to_bool( TestCase ):
    def setUp( self ):
        self.list_of_conbinations_to_bool = [
            ( 'false', False ), ( 'true', True ), ( '0', False ),
            ( '1', True ), ( object(), True ), ( 0, False ), ( 1, True ),
            ( 'Y', True ), ( 'y', True ), ( 'N', False ), ( 'n', False ),
        ]

    def test_the_list_of_combinations_should_return_all_the_expects( self ):
        for value, expected in self.list_of_conbinations_to_bool:
            result = to_bool( value )
            self.assertEqual(
                result, expected,
                "se esperaba {} con el valor de {} pero se obtubo {}".format(
                    expected, value, result ) )



from chibi.file.snippets import mkdir, exists
from tests.snippet.files import Test_with_files


class Test_mkdir( Test_with_files ):
    def test_create_a_new_dir_and_that_exists_should_be_ok_in_default( self ):
        mkdir( self.dirs[0] )

    def test_create_a_new_dir_and_that_exists_should_be_should_raise( self ):
        with self.assertRaises( OSError ):
            mkdir( self.dirs[0], False )

    def test_print_the_directory_name_when_is_verbose( self ):
        with self.assertLogs( level='INFO' ) as cm:
            new_dir = self.dirs[0] + 'asdf'
            if exists( new_dir ):
                self.fail(
                    "el directiorio {} ya existe usar"
                    "otro nombre".format( new_dir ) )

            mkdir( new_dir, verbose=True )
            if not exists( new_dir ):
                self.fail(
                    "no se creo el directorio {} ".format( new_dir ) )

        message_print = cm.output[0]
        self.assertRegex( message_print, r".+'{}'".format( new_dir ) )

    def test_if_the_directory_no_exists_should_create( self ):
        new_dir = self.dirs[0] + 'asdf'
        if exists( new_dir ):
            self.fail(
                "el directiorio {} ya existe usar"
                "otro nombre".format( new_dir ) )
        mkdir( new_dir, verbose=False )
        if not exists( new_dir ):
            self.fail(
                "no se creo el directorio {} ".format( new_dir ) )



class Chibi_chain( list ):
    def __init__( self, *args, next_obj=None, retrieve_next=None, **kw ):
        super().__init__( *args, **kw )
        self.next_obj = next_obj
        self.retrieve_next = retrieve_next
        self._validate_next_obj()

    def __iter__( self ):
        iterador = super().__iter__()
        for i in iterador:
            yield i
        if self.next_obj:
            next_obj = self.retrieve_next( self.next_obj )
            self.next_obj = None
            next_obj_iter = iter( next_obj )
            for i in next_obj_iter:
                self.append( i )
                yield i

    def _validate_next_obj( self ):
        if self.next_obj is not None and self.retrieve_next is None:
            raise NotImplementedError(
                "se nesesita una definicion de funcion para obtener "
                "los siguientes items" )

        if self.next_obj is not None and not callable( self.retrieve_next ):
            raise NotImplementedError(
                "retrieve_next deberia de ser llamable" )
        return True



from unittest import TestCase
from chibi.file.temp import Chibi_temp_path
from chibi.file.other import Chibi_conf


content = """
; Start a new pool named 'www'.
[www]

; Unix user/group of processes
; Note: The user is mandatory. If the group is not set, the
; default user's group
;       will be used.
; RPM: apache Choosed to be able to access some dir as httpd
user = apache
; RPM: Keep a group allowed to write in log dir.
group = apache

; The address on which to accept FastCGI requests.
; Valid syntaxes are:
;   'ip.add.re.ss:port'    - to listen on a TCP socket to a specific
;                            a specific port;
;   '[ip:6:addr:ess]:port' - to listen on a TCP socket to a specific
;                            a specific port;
;   'port'                 - to listen on a TCP socket to all addresses
;                            (IPv6 and IPv4-mapped) on a specific port;
;   '/path/to/unix/socket' - to listen on a unix socket.
; Note: This value is mandatory.
listen = 127.0.0.1:9000

; Set listen(2) backlog.
; Default Value: 511 (-1 on FreeBSD and OpenBSD)
;listen.backlog = 511

; Set permissions for unix socket, if one is used. In Linux, read/write
; permissions must be set in order to allow connections from a web server. Many
; BSD-derived systems allow connections regardless of permissions.
; Default Values: user and group are set as the running user
;                 mode is set to 0660
;listen.owner = nobody
;listen.group = nobody
;listen.mode = 0660
; When POSIX Access Control Lists are supported you can set them using
; these options, value is a comma separated list of user/group names.
; When set, listen.owner and listen.group are ignored
;listen.acl_users =
;listen.acl_groups =

; List of addresses (IPv4/IPv6) of FastCGI clients which are
; Equivalent to the FCGI_WEB_SERVER_ADDRS environment variable
; PHP FCGI (5.2.2+). Makes sense only with a tcp listening socket
; must be separated by a comma. If this value is left blank, connections
; accepted from any ip address.
; Default Value: any
listen.allowed_clients = 127.0.0.1
"""


class Test_chibi_conf( TestCase ):
    def setUp( self ):
        self.folder = Chibi_temp_path()
        self.file_service = self.folder.temp_file( extension='conf' )
        with open( self.file_service, 'w' ) as f:
            f.write( content )

    def test_should_be_a_dict( self ):
        conf = Chibi_conf( self.file_service )
        result = conf.read()
        self.assertIsInstance( result, dict )

    def test_should_have_expected_data( self ):
        conf = Chibi_conf( self.file_service )
        result = conf.read()
        expected = {
            'www': {
                'group': 'apache',
                'listen': '127.0.0.1:9000',
                'listen.allowed_clients': '127.0.0.1',
                'user': 'apache',
            }
        }
        self.assertEqual( result, expected )

    def test_when_change_variable_should_work( self ):
        conf = Chibi_conf( self.file_service )
        result = conf.read()
        result[ 'www' ][ 'user' ] = 'qwert'
        conf.write( result )
        result_2 = conf.read()
        expected = {
            'www': {
                'group': 'apache',
                'listen': '127.0.0.1:9000',
                'listen.allowed_clients': '127.0.0.1',
                'user': 'qwert',
            }
        }
        self.assertEqual( result_2, expected )



from unittest import TestCase

from chibi.snippet.xml import guaranteed_list, compress_dummy_list


class test_guaranteed_list(TestCase):

    def setUp( self ):
        self.example = {
            'args': 'nmap -oX - -sn 200.200.200.200/24',
            'debugging': {'level': '0'},
            'host': {
                'address': {'addr': '200.200.200.200', 'addrtype': 'ipv4'},
                'hostnames': None,
                'status': {'reason': 'conn-refused', 'reason_ttl': '0',
                           'state': 'up'},
                'times': {'rttvar': '3768', 'srtt': '204', 'to': '100000'}},
            'runstats': {
                'finished': {
                    'elapsed': '6.82', 'exit': 'success',
                    'summary': (
                        'Nmap done at Thu May  2 15:24:38 '
                        '2019; 256 IP addresses (1 host up) '
                        'scanned in 6.82 seconds' ),
                    'time': '1556828678',
                    'timestr': 'Thu May  2 15:24:38 2019'},
                'hosts': {'down': '255', 'total': '256', 'up': '1'}},
            'scanner': 'nmap', 'start': '1556828671',
            'startstr': 'Thu May  2 15:24:31 2019', 'verbose': {'level': '0'},
            'version': '7.70', 'xmloutputversion': '1.04'}

        self.expected = {
            'args': 'nmap -oX - -sn 200.200.200.200/24',
            'debugging': {'level': '0'},
            'host': [ {
                'address': {'addr': '200.200.200.200', 'addrtype': 'ipv4'},
                'hostnames': None,
                'status': {'reason': 'conn-refused', 'reason_ttl': '0',
                           'state': 'up'},
                'times': {'rttvar': '3768', 'srtt': '204', 'to': '100000'}} ],
            'runstats': {
                'finished': {
                    'elapsed': '6.82', 'exit': 'success',
                    'summary': (
                        'Nmap done at Thu May  2 15:24:38 '
                        '2019; 256 IP addresses (1 host up) '
                        'scanned in 6.82 seconds' ),
                    'time': '1556828678',
                    'timestr': 'Thu May  2 15:24:38 2019'},
                'hosts': {'down': '255', 'total': '256', 'up': '1'}},
            'scanner': 'nmap', 'start': '1556828671',
            'startstr': 'Thu May  2 15:24:31 2019', 'verbose': {'level': '0'},
            'version': '7.70', 'xmloutputversion': '1.04'}

    def test_should_convert_host_in_list( self ):
        result = guaranteed_list( self.example, 'host' )
        self.assertEqual( self.expected, result )


class test_compress_dummy_list(TestCase):

    def setUp( self ):
        self.example = {
            'regions': { 'region': 'asdf' },
            'attrs': { 'attr': { 'asdf': 'asdf' } },
            'lists': [ '', [], [ { 'regions': { 'region': 'qq' } } ] ],
            'list': [
                '',
                [ { 'regions': { 'region': 'qq' } } ],
                [
                    {
                        'regions': {
                            'region': [ { 'asdfs': { 'asdf': 1 } } ]
                        }
                    },
                ],
            ],
        }

        self.expected = {
            'attrs': {'asdf': 'asdf'},
            'list': [
                '',
                [ { 'regions': 'qq' } ],
                [ { 'regions': [ { 'asdfs': 1 } ] } ] ],
            'lists': [ '', [], [ { 'regions': 'qq' } ] ],
            'regions': 'asdf', }

    def test_should_convert_host_in_list( self ):
        result = compress_dummy_list( self.example )
        self.assertEqual( self.expected, result )



import glob
import distutils.dir_util
import logging
import os
import shutil
import re

import magic


logger = logging.getLogger( "chibi.file.chibi_path" )


class Chibi_path( str ):
    def __new__( cls, *args, chibi_file_class=None, **kw ):
        args_2 = []
        for a in args:
            if '~' in a:
                a = os.path.expanduser( a )
            args_2.append( a )
        result = str.__new__( cls, *args_2, **kw )
        result._chibi_file_class = chibi_file_class
        return result

    def __add__( self, other ):
        """
        une el path con otro path o una cadena

        Parameters
        ==========
        other: str or Chibi_path

        Returns
        =======
        Chibi_path
        """
        if isinstance( other, self.__class__ ):
            if self.is_a_file:
                return self.dir_name + other

            return type( self )( os.path.join( str( self ), str( other ) ) )
        if isinstance( other, str ):
            return self + self.__class__( other )

    def __eq__( self, other ):
        if isinstance( other, Chibi_path ):
            return str( self ) == str( other )
        if isinstance( other, str ):
            return str( self ) == other
        return False

    def __hash__( self ):
        return hash( str( self ) )

    def __contains__( self, other ):
        if isinstance( other, Chibi_path ):
            return other.startswith( self )
        else:
            return super().__contains__( other )

    @property
    def is_a_folder( self ):
        """
        es una carpeta
        """
        return os.path.isdir( self )

    @property
    def is_a_file( self ):
        """
        es un archivo
        """
        from chibi.file.snippets import is_a_file
        return is_a_file( self )

    @property
    def is_glob( self ):
        return glob.has_magic( self )

    @property
    def dir_name( self ):
        """
        regresa la carpeta padre
        """
        from chibi.file.snippets import file_dir
        return type( self )( os.path.dirname( str( self ) ) )
        return self.__class__( file_dir( self ) )

    @property
    def base_name( self ):
        """
        regresa el nombre del archivo o de la carpeta
        """
        return Chibi_path( os.path.basename( self ) )

    @property
    def file_name( self ):
        """
        regresa el nombre del archivo sin la extencion
        """
        file_name, ext = os.path.splitext( self.base_name )
        return file_name

    def open( self, chibi_file_class=None ):
        """
        abre el archivo usando un chibi file
        """
        if self.is_a_folder:
            raise NotImplementedError
        if chibi_file_class is None:
            if self._chibi_file_class is None:
                from .file import Chibi_file
                chibi_file_class = Chibi_file
            else:
                chibi_file_class = self._chibi_file_class
        return chibi_file_class( self )

    def relative_to( self, root ):
        from .snippets import get_relative_path
        return type( self )( get_relative_path( self, root=root ) )

    def mkdir( self, **kw ):
        """
        crea una carpeta en la direcion del chibi path
        """
        try:
            os.makedirs( self )
            logger.info( "se creo el directorio '{}'".format( self ) )
        except OSError:
            pass
        if kw:
            logger.warning(
                "mkdir de chibi path recibio parametros {}".format( kw ) )

    def move( self, dest ):
        """
        move the chibi path al destino
        """
        dest = Chibi_path( dest )
        if self.is_a_file:
            if dest.is_a_folder:
                dest += self.base_name
            shutil.move( str( self ), str( dest ) )
            logger.info( "{} -> {}".format( self, dest ) )
        elif self.is_a_folder:
            shutil.move( str( self ), str( dest ) )
            logger.info( "{} -> {}".format( self, dest ) )
        elif self.is_glob:
            if not dest.exists:
                dest.mkdir()
            if dest.is_a_folder:
                return [
                    f.move( dest + f.base_name ) for f in self.expand ]
            else:
                raise NotImplementedError(
                    "el destino no es un folder y la src "
                    "es un glob '{self}'" )

    def copy( self, dest, **kw ):
        """
        copia el archivo o carpeta al destino
        """
        from.snippets import copy
        dest = Chibi_path( dest )
        if self.is_a_file:
            if dest.is_a_folder:
                dest += self.base_name
            copy( self, dest, **kw )
            return Chibi_path( dest )
        elif self.is_a_folder:
            if dest.is_a_file:
                raise NotImplementedError(
                    "no se puede copiar un folder dentro de un archivo" )
            distutils.dir_util.copy_tree( str( self ), str( dest ) )
            return Chibi_path( dest )
        elif self.is_glob:
            if not dest.exists:
                dest.mkdir()
            if dest.is_a_folder:
                return [ f.copy( dest + f.base_name ) for f in self.expand ]
            else:
                raise NotImplementedError(
                    "el destino no es un folder y la src "
                    "es un glob '{self}'" )
        if not self.exists:
            raise OSError( f"the file '{self}' not exists" )
        else:
            raise NotImplementedError(
                f"no esta implementado el copy si "
                f"no es un archivo o folder '{self}'" )

    def delete( self ):
        """
        elimina el archivo o la carpeta
        """
        from.snippets import delete
        delete( str( self ) )
        logger.info( 'delete "{}"'.format( self ) )

    def chown(
            self, verbose=True, user_name=None, group_name=None,
            recursive=False ):
        """
        cambia el duano del archivo o carpeta
        """
        from chibi.file.snippets import chown
        chown(
            self, user_name=user_name, group_name=group_name,
            recursive=recursive )

    def chmod( self, mod ):
        """
        cambia los attributos del archivo o carpeta
        """
        os.chmod( str( self ), mod )

    @property
    def properties( self ):
        from chibi.file.snippets import stat

        prop = stat( self )
        prop.mime = magic.Magic( mime=True ).from_file( self )
        prop.extension = os.path.splitext( self )[1][1:]
        return prop

    @property
    def extension( self ):
        """
        regresa la extencion del archivo
        """
        if self.is_a_file:
            return self.properties.extension
        else:
            raise NotImplementedError

    def replace_extensions( self, *extensions ):
        """
        cambia la extencion del archivo
        """
        file_name, ext = os.path.splitext( self )
        extensions = ".".join( extensions )
        file_name = ".".join( ( file_name, extensions ) )
        return type( self )( file_name )

    def add_extensions( self, *extensions ):
        """
        agrega mas extenciones
        """
        file_name, ext = os.path.splitext( self )
        extensions = ".".join( extensions )
        file_name = ".".join( ( file_name, ext + extensions ) )
        return type( self )( file_name )

    def ls( self, dirs=True, files=True ):
        """
        regresa un generador con el listado de archivos y carpetas
        """
        from .snippets import ls, ls_only_dir, ls_only_files
        if dirs and files:
            return ls( self )
        elif dirs and not files:
            return ls_only_dir( self )
        elif not dirs and files:
            return ls_only_files( self )
        else:
            raise NotImplementedError

    def find( self, search_term=".*", dirs=True, files=True ):
        """
        busca archivos y carpetas usando una exprecion regular
        """
        if self.is_a_file:
            raise NotImplementedError(
                "no esta implementa buscar en un archivo" )
        from .snippets import find, find_only_files, find_only_folders
        if dirs and files:
            return find( self, search_term )
        elif dirs and not files:
            return find_only_folders( self, search_term )
        elif not dirs and files:
            return find_only_files( self, search_term )
        else:
            raise NotImplementedError

    @property
    def exists( self ):
        """
        revisa si el archivo o directorio existe

        Returns
        =======
        bool
        """
        from .snippets import exists
        return exists( str( self ) )

    def replace( self, *args, **kw ):
        return Chibi_path( super().replace( *args, **kw ) )

    def made_safe( self ):
        return Chibi_path( re.sub( r'[<>:"|?*]', '', str( self ) ) )

    @classmethod
    def current_dir( cls ):
        """
        regresa el directorio actual de trabajo

        Returns
        =======
        py:class:`chibi.file.Chibi_path`
        """
        return Chibi_path( os.getcwd() )

    @property
    def expand( self ):
        if self.is_glob:
            return ( type( self )( f ) for f in glob.iglob( self ) )
        else:
            raise NotImplementedError( "no se que deberia de hacer" )

    def __enter__( self ):
        return self.open()

    def __exit__( self, exc_type, exc_value, traceback ):
        pass

    def touch( self ):
        if not self.exists or self.is_a_file:
            self.open().touch()
        elif self.is_a_folder:
            raise NotADirectoryError(
                f"no implementado touch a un folder '{self}'" )
        else:
            raise NotADirectoryError(
                f"no implementado touch cuando no es un "
                f"archivo o folder'{self}'" )

    @property
    def inflate( self ):
        if '~' in self:
            return os.path.expanduser( self )
        else:
            return os.path.abspath( self )



from chibi.units.base import Unit
from unittest import TestCase


class Test_unit( TestCase ):
    def setUp( self ):
        self.unit = Unit( 10 )

    def test_should_print_the_value_when_is_str( self ):
        self.assertIn( '10', str( self.unit ) )

    def test_when_add_a_int_should_work( self ):
        r = 10 + self.unit
        self.assertEqual( r.value, 20 )
        r = self.unit + 10
        self.assertEqual( r.value, 20 )

    def test_when_add_a_float_should_work( self ):
        r = 10.10 + self.unit
        self.assertEqual( r.value, 20.1 )
        r = self.unit + 10.1
        self.assertEqual( r.value, 20.1 )

    def test_when_sub_a_int_should_work( self ):
        r = 10 - self.unit
        self.assertEqual( r.value, 0 )
        r = self.unit - 10
        self.assertEqual( r.value, 0 )

    def test_when_sub_a_float_should_work( self ):
        r = 10.10 - self.unit
        self.assertAlmostEqual( r.value, -0.1, delta=0.01 )
        r = self.unit - 10.10
        self.assertAlmostEqual( r.value, -0.1, delta=0.01 )

    def test_when_mul_a_int_should_work( self ):
        r = 10 * self.unit
        self.assertEqual( r.value, 100 )
        r = self.unit * 10
        self.assertEqual( r.value, 100 )

    def test_when_mul_a_float_should_work( self ):
        r = 10.1 * self.unit
        self.assertEqual( r.value, 101.0 )
        r = self.unit * 10.1
        self.assertEqual( r.value, 101.0 )

    def test_when_div_a_int_should_work( self ):
        r = 10 / self.unit
        self.assertEqual( r.value, 1 )
        r = self.unit / 10
        self.assertEqual( r.value, 1 )

    def test_when_div_a_float_should_work( self ):
        r = 10.10 / self.unit
        self.assertAlmostEqual( r.value, 0.99, delta=0.001 )
        r = self.unit / 10.10
        self.assertAlmostEqual( r.value, 0.99, delta=0.001 )

    def test_when_div_int_a_int_should_work( self ):
        r = 10 // self.unit
        self.assertEqual( r.value, 1 )
        r = self.unit // 10
        self.assertEqual( r.value, 1 )

    def test_when_div_int_a_float_should_work( self ):
        r = 10.10 // self.unit
        self.assertEqual( r.value, 0 )
        r = self.unit // 10.10
        self.assertEqual( r.value, 0 )

    def test_when_pow_a_int_should_work( self ):
        r = 10 ** self.unit
        self.assertEqual( r.value, 10000000000 )
        r = self.unit ** 10
        self.assertEqual( r.value, 10000000000 )

    def test_when_pow_float_a_float_should_work( self ):
        r = 10.10 ** self.unit
        self.assertEqual( r.value, 12589254117.941662 )
        r = self.unit ** 10.10
        self.assertEqual( r.value, 12589254117.941662 )
